import 'dart:async';
import 'package:_private_chat/firestore_resouce/instances.dart';
import 'package:_private_chat/options.dart';
import 'package:_private_core/_private_core.dart';
import 'package:_private_core_network/network_resources/resources.dart';
import 'package:bloc_concurrency/bloc_concurrency.dart';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:_private_chat/chat_detail/chat_detail_screen.dart';
import 'package:_private_chat/chat_detail/chat_helper.dart';
import 'package:_private_chat/utils/utils.dart';
import 'package:_private_chat/constant.dart';

import '../../connection_status/bloc/connection_status_bloc.dart';

part 'chat_list_event.dart';
part 'chat_list_state.dart';

String? chatGetFullNameUser(CommonUserModel? data) => data?.displayName;

String? chatGetThumbnailUser(CommonUserModel? data) => data?.thumbnail;

String? chatGetFirebaseIdUser(CommonUserModel? data) => data?.firebaseId;

ChatListBloc get chatListBloc => appFindInstance<ChatListBloc>();

class ChatListConversation {
  int timestamp;
  ChatConversationType conversationType;
  String conversationId;
  Map data = {};
  List userIds = [];
  int unread;
  Map<String, dynamic> lastMessage = {};
  bool isEmptyMessages;
  bool isBot;

  ChatListConversation({
    this.timestamp = 0,
    this.isBot = false,
    this.conversationId = '',
    this.conversationType = ChatConversationType.solo,
    this.unread = 0,
    this.isEmptyMessages = false,
  });
}

class ChatListBloc extends Bloc<ChatListEvent, ChatListState> {
  StreamSubscription? _converstationsSub;
  final List<StreamSubscription> _messageSubs = [];

  ChatListBloc() : super(ChatListState(conversations: [], usersByKeys: {})) {
    on<ChatListInitEvent>(
      _onInitial,
      transformer: restartable(),
    );
    on<ChatListReadItemEvent>(_onReadItemEvent);
    on<ChatListDisposeEvent>(_onDisposeEvent);
    on<ChatListRemoveEvent>(onRemoveEvent);
    on<ChatUpdateListConversationEvent>(_onUpdateConversation);
    on<LoadUserInfoConversationEvent>(_onLoadUserInfo);
  }

  _onUpdateConversation(
      ChatUpdateListConversationEvent event, Emitter<ChatListState> emit) {
    emit(state.update(conversations: event.conversations));
  }

  onRemoveEvent(ChatListRemoveEvent event, Emitter<ChatListState> emit) {
    ChatHelper.removeConversation(event.conversationId);
    state.conversations
        .removeWhere((e) => e.conversationId == event.conversationId);
    emit(state.update());
  }

  _onReadItemEvent(ChatListReadItemEvent event, Emitter<ChatListState> emit) {
    if (event.c != null) {
      if (event.c!.isBot) {
        appPushNamed(
          event.context,
          ChatDetailScreen.route,
          arguments: ChatDetailParams(
            event.c!.conversationType,
            isBot: true,
          ),
        );
      } else {
        if (event.c!.conversationType == ChatConversationType.solo) {
          String firebaseId = event.c!.userIds.first[kDbFIREBASEID];
          CommonUserModel? user = state.getUserByKeys(firebaseId);
          String? fullname = chatGetFullNameUser(user);
          String? thumbnail = chatGetThumbnailUser(user);
          appPushNamed(event.context, ChatDetailScreen.route,
              arguments: ChatDetailParams.solo(
                soloFirebaseId: firebaseId,
                soloUser: user,
                name: fullname,
                thumbnail: thumbnail,
              ));
        } else {
          appPushNamed(
            event.context,
            ChatDetailScreen.route,
            arguments: ChatDetailParams.group(
              conversationId: event.c!.conversationId,
              name: event.c!.data[kDbGROUPNAME],
              thumbnail: event.c!.data[kDbTHUMBNAIL],
            ),
          );
        }
      }
    } else if (event.firebaseId != null) {
      appPushNamed(
        event.context,
        ChatDetailScreen.route,
        arguments: ChatDetailParams.solo(
          soloFirebaseId: event.firebaseId,
          name: event.displayName,
          thumbnail: event.thumbnail,
        ),
      );
    }
  }

  _onDisposeEvent(ChatListDisposeEvent event, Emitter<ChatListState> emit) {
    _converstationsSub?.cancel();
    for (var e in _messageSubs) {
      e.cancel();
    }
    _messageSubs.clear();
    chatConnectionStatusBloc.add(ConnectionStatusDisposeEvent());
    add(ChatListInitEvent(isJustForReset: true));
    if (event.alsoResetConversations) {
      emit(state.update(conversations: []));
    }
  }

  _onLoadUserInfo(
      LoadUserInfoConversationEvent event, Emitter<ChatListState> emit) async {
    if (!isLoggedFirebaseAuth()) return;
    Map<String, CommonUserModel> usersByKeys = state.usersByKeys;

    for (var firebaseId in event.firebaseIds) {
      if (firebaseId != '' &&
          firebaseId != loggedFirebaseId &&
          (state.usersByKeys[firebaseId] == null ||
              state.usersByKeys[firebaseId]?.username == "anonymous")) {
        usersByKeys[firebaseId] =
            await getInfoByFirebaseId(firebaseId, isAnonymously: false);
      }
    }

    emit(state.update(usersByKeys: usersByKeys));
  }

  _onInitial(ChatListInitEvent event, Emitter<ChatListState> emit) async {
    if (!isLoggedFirebaseAuth() || event.isJustForReset) return;
    bool showConverstaion(docData) {
      List users = docData[kDbMEMBERSDETAIL]
          .where((e) => e[kDbFIREBASEID] != loggedFirebaseId)
          .toList();
      return (event.conditionInitialize?.call(docData) ?? true) &&
          users.isNotEmpty &&
          !users.any((e) => e == null);
    }

    try {
      for (var e in _messageSubs) {
        e.cancel();
      }
      _messageSubs.clear();
      emit(state.update(
          isLoading: true, isLoadingInitialize: event.isLoadingInitialize));
      QuerySnapshot q = await colCONVERSATIONS
          .where(kDbMEMBERSFIREBASEID, arrayContains: loggedFirebaseId)
          .orderBy(kDblastMessageTime, descending: true)
          .get();
      List<Map> requestUsers = [];
      List<ChatListConversation> cons = [];
      if (chatOptions.botEnable) {
        cons.add(ChatListConversation(isBot: true));
      }

      for (var d in q.docs) {
        Map docData = d.data() as Map;
        if (showConverstaion(docData)) {
          int timestamp = docData[kDbMEMBERSDETAIL]
                  .any((e) => e[kDbFIREBASEID] == loggedFirebaseId)
              ? docData[kDbMEMBERSDETAIL].firstWhere((e) =>
                      e[kDbFIREBASEID] == loggedFirebaseId)[kDbtimestamp] ??
                  0
              : 0;
          ChatListConversation conversation = ChatListConversation()
            ..conversationId = d.id
            ..userIds = docData[kDbMEMBERSDETAIL]
                .where((e) => e[kDbFIREBASEID] != loggedFirebaseId)
                .toList()
            ..conversationType =
                ChatConversationType.values.byName(docData[kDbCONVERSATIONTYPE])
            ..data = docData;
          requestUsers.addAll((docData[kDbMEMBERSDETAIL] as List)
              .map<Map>((e) => e as Map)
              .toList());

          if (docData[kDblastMessageTime] != null &&
              docData[kDblastMessageTime] != 0) {
            conversation.timestamp = docData[kDblastMessageTime];
            conversation.unread = await countUnread(d.id, timestamp);
            conversation.lastMessage = await fetchLatestMessage(d.id,
                conversation.data[kDbENCRYPTKEY], docData[kDblastMessageTime]);
            cons.add(conversation);
          } else {
            conversation.unread = 0;
            conversation.isEmptyMessages = true;
            conversation.lastMessage = {};
            cons.add(conversation);
          }
        }
      }

      emit(state.update(
        isLoading: false,
        isLoadingInitialize: false,
        conversations: cons,
      ));

      _converstationsSub?.cancel();
      _converstationsSub = colCONVERSATIONS
          .where(kDbMEMBERSFIREBASEID, arrayContains: loggedFirebaseId)
          .snapshots()
          .listen((snapshot) async {
        snapshot.docChanges.forEach((e) {
          print('snapshot.docChanges: ${e.type} ${e.doc.data()}');
        });
        if (snapshot.docs.length != state.conversations.length) {
          _converstationsSub?.cancel();
          add(ChatListInitEvent(isLoadingInitialize: false));
        }
      });

      for (var d in q.docs) {
        if (!showConverstaion(d.data())) return;
        _messageSubs.add(
            colCONVERSATIONS.doc(d.id).snapshots().listen((snapshot) async {
          Map? docData = snapshot.data();
          if (docData == null) return;

          //Handle callback new message
          if (state.conversations.any((e) {
            if (e.conversationId != d.id) return false;
            if (docData[kDblatestMessage]?[kDbfrom] == loggedFirebaseId) {
              return false;
            }
            return (e.data[kDblastMessageTime] ?? 0) <
                (docData[kDblastMessageTime] ?? 0);
          })) {
            chatOptions.onNewMessageCallback
                ?.call(d.id, docData[kDblatestMessage]);
          }

          int timestamp = docData[kDbMEMBERSDETAIL]
                  .any((e) => e[kDbFIREBASEID] == loggedFirebaseId)
              ? docData[kDbMEMBERSDETAIL].firstWhere((e) =>
                      e[kDbFIREBASEID] == loggedFirebaseId)[kDbtimestamp] ??
                  0
              : 0;
          ChatListConversation conversation = state.conversations.firstWhere(
              (e) => e.conversationId == d.id,
              orElse: () => ChatListConversation())
            ..conversationId = d.id
            ..userIds = docData[kDbMEMBERSDETAIL]
                .where((e) => e[kDbFIREBASEID] != loggedFirebaseId)
                .toList()
            ..conversationType =
                ChatConversationType.values.byName(docData[kDbCONVERSATIONTYPE])
            ..data = docData;

          if (docData[kDblastMessageTime] != null &&
              docData[kDblastMessageTime] != 0) {
            conversation.unread = await countUnread(d.id, timestamp);
            if (conversation.timestamp != docData[kDblastMessageTime]) {
              //Last message
              conversation.timestamp = docData[kDblastMessageTime];
              conversation.lastMessage = await fetchLatestMessage(
                  d.id,
                  conversation.data[kDbENCRYPTKEY],
                  docData[kDblastMessageTime]);
              conversation.isEmptyMessages = false;
            }
          } else {
            conversation.unread = 0;
            conversation.isEmptyMessages = true;
            conversation.lastMessage = {};
          }

          List<ChatListConversation> conversations =
              List<ChatListConversation>.from(state.conversations
                      .where((e) => e.conversationId != d.id)) +
                  [conversation];
          conversations.sort((a, b) => b.timestamp > a.timestamp ? 1 : -1);
          add(ChatUpdateListConversationEvent(conversations));
        }));
      }

      // _getConnectionStatusUsers(requestUsers);
      emit(state.update(usersByKeys: await _getUsersByKeys(requestUsers)));
    } catch (_) {
      appDebugPrint("Error ChatListInitEvent: $_");
    }
  }

  // void _getConnectionStatusUsers(List<Map> requestUsers) async {
  // List<String> users = requestUsers
  //     .map((e) => e[kDbFIREBASEID]?.toString() ?? "")
  //     .where((e) => e != '')
  //     .toList();
  // if (users.isNotEmpty) {
  //   chatConnectionStatusBloc.add(ConnectionStatusInitEvent(users));
  // }
  // }

  Future<Map<String, CommonUserModel>> _getUsersByKeys(
      List<Map> requestUsers) async {
    Map<String, CommonUserModel> usersByKeys = state.usersByKeys;
    for (var e in requestUsers) {
      String firebaseId = e[kDbFIREBASEID] ?? "";
      bool isAnonymously = e[kDbisAnonymously] ?? false;
      if (firebaseId != '' &&
          firebaseId != loggedFirebaseId &&
          (state.usersByKeys[firebaseId] == null ||
              state.usersByKeys[firebaseId]?.username == "anonymous")) {
        usersByKeys[firebaseId] =
            await getInfoByFirebaseId(firebaseId, isAnonymously: isAnonymously);
      }
    }
    return usersByKeys;
  }

  Future<int> countUnread(String converstationId, int timestamp) async {
    return (await colCONVERSATIONS
                .doc(converstationId)
                .collection(kCollectionMESSAGES)
                .where(kDbtimestamp, isGreaterThan: timestamp)
                .count()
                .get())
            .count ??
        0;
  }

  Future<Map<String, dynamic>> fetchLatestMessage(
    String converstationId,
    String encryptKey,
    int messageId,
  ) async {
    //Last message
    DocumentSnapshot docLastMessage = await colCONVERSATIONS
        .doc(converstationId)
        .collection(kCollectionMESSAGES)
        .doc('$messageId')
        .get();
    Map<String, dynamic> data = {};
    if (docLastMessage.data() != null) {
      data = docLastMessage.data() as Map<String, dynamic>;
      data[kDbCONTENT] = ChatHelper.decrypt(encryptKey, data[kDbCONTENT]);
    }
    return data;
  }
}
